# -*- coding: UTF-8 -*-
import sqlite3
import time
import logging
import os
from src.jcci.sql import create_tables

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.DEBUG)

class SqliteHelper(object):
    def __init__(self, db_path):
        self.db_path = db_path

    def connect(self):
        try:
            if not os.path.exists(os.path.dirname(self.db_path)):
                os.makedirs(os.path.dirname(self.db_path))
            if not os.path.exists(self.db_path):
                db_file = open(self.db_path, "x")
                db_file.close()
                conn = sqlite3.connect(self.db_path)
                # 执行创建表的SQL语句
                try:
                    conn.cursor().executescript(create_tables)
                    logging.info("Table created successfully")
                except sqlite3.Error as e:
                    logging.error(f"Error creating table: {e}")
            else:
                conn = sqlite3.connect(self.db_path)
        except Exception as e:
            logging.error(f'connect fail {e}')
            time.sleep(1)
            conn = sqlite3.connect(self.db_path)
        return conn

    def add_project(self, project_name, git_url, branch, commit_or_branch_new, commit_or_branch_old):
        try:
            projects = self.select_data(f'SELECT * FROM project where project_name="{project_name}" and git_url="{git_url}" '
                                        f'and branch="{branch}" AND commit_or_branch_new="{commit_or_branch_new}" and commit_or_branch_old="{commit_or_branch_old}"')
            if projects:
                return projects[0]['project_id']
            conn = self.connect()
            c = conn.cursor()
            c.execute(f'INSERT INTO project '
                      f'(project_name, git_url, branch, commit_or_branch_new, commit_or_branch_old) '
                      f'VALUES("{project_name}", "{git_url}", "{branch}", "{commit_or_branch_new}", "{commit_or_branch_old}")')
            project_id = c.lastrowid
            conn.commit()
            conn.close()
            return project_id
        except Exception as e:
            logging.error(f'add_project fail')

    def add_class(self, filepath, access_modifier, class_type, class_name, package_name, extends_class, project_id, implements, annotations, documentation, is_controller, controller_base_url, commit_or_branch):
        try:
            class_list = self.select_data(f'SELECT * FROM class WHERE project_id={project_id} and package_name="{package_name}" and class_name="{class_name}" and commit_or_branch="{commit_or_branch}"')
            if class_list:
                return class_list[0]['class_id'], False
            conn = self.connect()
            c = conn.cursor()
            c.execute('INSERT INTO class (filepath, access_modifier, class_type, class_name, package_name, extends_class, project_id, implements, annotations, documentation, is_controller, controller_base_url, commit_or_branch) '
                      'VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)', (filepath, access_modifier, class_type, class_name, package_name, extends_class, project_id, implements, annotations, documentation, is_controller, controller_base_url, commit_or_branch))
            class_id = c.lastrowid
            conn.commit()
            conn.close()
            return class_id, True
        except Exception as e:
            logging.error(e)
            logging.error(f'add_class fail')

    def select_data(self, sql):
        try:
            conn = self.connect()
            c = conn.cursor()
            cursor = c.execute(sql)
            res = cursor.fetchall()
            columns = cursor.description
            field = [column_name[0] for column_name in columns]
            zip_data = [dict(zip(field, item)) for item in res]
            conn.close()
            return zip_data
        except Exception as e:
            logging.error(f'select_data fail')
            raise e

    def update_data(self, sql):
        try:
            conn = self.connect()
            c = conn.cursor()
            c.execute(sql)
            conn.commit()
            conn.close()
        except Exception as e:
            logging.error(f'select_data fail')

    def insert_data(self, table_name: str, data) -> bool:
        try:
            conn = self.connect()
            c = conn.cursor()
            if isinstance(data, list):
                for item in data:
                    keys = ",".join(list(item.keys()))
                    values = ",".join([f'''"{x.replace('"', '""').replace("'", "''")}"''' if isinstance(x, str) else f"'{x}'" for x in list(item.values())])
                    sql = f"INSERT INTO {table_name} ({keys}) VALUES ({values});"
                    c.execute(sql)
            elif isinstance(data, dict):
                keys = ",".join(list(data.keys()))
                values = ",".join([f"'{x}'" for x in list(data.values())])
                sql = f"INSERT INTO {table_name} ({keys}) VALUES ({values});"
                c.execute(sql)
            conn.commit()
            conn.close()
            return True
        except Exception as ex:
            logging.error(f"insert data error {ex}")
            return False



# -*- coding: UTF-8 -*-
import xml.etree.ElementTree as ET


class Mapper(object):
    def __init__(self, namespace, result_maps, sqls, statements):
        self.namespace = namespace
        self.result_maps = result_maps
        self.sqls = sqls
        self.statements = statements


class MapperElement(object):
    def __init__(self, id, type, start, end, content):
        self.id = id
        self.name = id
        self.type = type
        self.start = start
        self.end = end
        self.content = content
        self.diff_impact = None


class MapperStatement(MapperElement):
    def __init__(self, id, type, start_line, end_line, content, statement_tag, result_map, include_sql):
        super(MapperStatement, self).__init__(id, type, start_line, end_line, content)
        self.statement_tag = statement_tag
        self.result_map = result_map
        self.include_sql = include_sql


def parse(filepath):
    # 读取XML文件内容
    with open(filepath, "r", encoding="utf-8") as file:
        xml_content = file.read()

    # 解析XML文件
    tree = ET.ElementTree(ET.fromstring(xml_content))
    root = tree.getroot()

    # 获取namespace
    try:
        namespace = root.attrib["namespace"]
        if namespace is None:
            return None
    except:
        return None
    # 存储resultMap和每条语句的id以及对应的起始行号和截止行号
    result_map_info = []
    sql_info = []
    statement_info = []

    # 获取resultMap的id以及起始行号和截止行号
    for element in root.findall(".//resultMap"):
        result_map_id = element.attrib["id"]
        start_line = 0
        end_line = 0
        for i, line in enumerate(xml_content.splitlines(), start=1):
            if line.strip().startswith('<resultMap') and f'id="{result_map_id}"' in line:
                start_line = i
            if f'</resultMap>' in line and start_line != 0:
                end_line = i
                break
        content = xml_content.splitlines()[start_line - 1: end_line]
        result_map_info.append(MapperElement(result_map_id, 'resultMap', start_line, end_line, content))

    # 获取resultMap的id以及起始行号和截止行号
    for sql_element in root.findall(".//sql"):
        sql_id = sql_element.attrib["id"]
        start_line = 0
        end_line = 0
        for i, line in enumerate(xml_content.splitlines(), start=1):
            if line.strip().startswith('<sql') and f'id="{sql_id}"' in line:
                start_line = i
            if f'</sql>' in line and start_line != 0:
                end_line = i
                break
        content = xml_content.splitlines()[start_line - 1: end_line]
        sql_info.append(MapperElement(sql_id, 'sql', start_line, end_line, content))

    # 获取每条语句的id以及起始行号和截止行号
    statements = root.findall(".//select") + root.findall(".//insert") + root.findall(".//update") + root.findall(".//delete")
    for statement_element in statements:
        statement_id = statement_element.attrib["id"]
        start_line = 0
        end_line = 0
        result_map = None
        include_sql = None
        for i, line in enumerate(xml_content.splitlines(), start=1):
            if f'<{statement_element.tag} id="{statement_id}"' in line:
                start_line = i
            if f'resultMap="' in line and start_line != 0:
                result_map = line.split('resultMap="')[1].split('"')[0]
            if line.strip().startswith('<include') and start_line != 0:
                include_sql = line.split('refid="')[1].split('"')[0]
            if f'</{statement_element.tag}>' in line and start_line != 0:
                end_line = i
                break
        content = xml_content.splitlines()[start_line - 1: end_line]
        statement_info.append(MapperStatement(statement_id, 'statement', start_line, end_line, content, statement_element.tag, result_map, include_sql))

    return Mapper(namespace, result_map_info, sql_info, statement_info)





ENTITY = 'entity'
RETURN_TYPE = 'return_type'
PARAMETERS = 'parameters'
BODY = 'body'
METHODS = 'methods'
FIELDS = 'fields'

PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN = 'unknown'
#

#
NODE_TYPE_CLASS = 'class'
NODE_TYPE_METHOD = 'method'
NODE_TYPE_FIELD = 'field'
NODE_TYPE_MAPPER = 'mapper'
NODE_TYPE_MAPPER_SQL = 'sql'
NODE_TYPE_MAPPER_RESULT_MAP = 'resultMap'
NODE_TYPE_MAPPER_STATEMENT = 'statement'

DIFF_TYPE_CHANGED = 'changed'
DIFF_TYPE_IMPACTED = 'impacted'

JAVA_BASIC_TYPE = ['string', 'int', 'boolean', 'long', 'byte', 'short', 'float', 'double', 'char']



import logging
import json
import javalang
from src.jcci.database import SqliteHelper
from src.jcci.constant import ENTITY, RETURN_TYPE, PARAMETERS, BODY, METHODS, FIELDS, PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN, JAVA_BASIC_TYPE

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.DEBUG)

class JavaParse(object):
    def __init__(self, db_path, project_id):
        self.project_id = project_id
        self.sqlite = SqliteHelper(db_path)

    def _handle_extends(self, extends, import_list: list, package_name):
        if isinstance(extends, list):
            extends_package_class_list = []
            extends_class = extends[0].name
            extends_package_class = self._get_extends_class_full_package(extends_class, import_list, package_name)
            extends_package_class_list.append(extends_package_class)
            if 'arguments' in extends[0].attrs and extends[0].arguments:
                extends_arguments = extends[0].arguments
                extends_argument_classes = []
                for extends_argument in extends_arguments:
                    if type(extends_argument) == javalang.tree.TypeArgument:
                        extends_argument_class = extends_argument.type.name
                    else:
                        extends_argument_class = extends_argument.name
                    extends_argument_package_class = self._get_extends_class_full_package(extends_argument_class, import_list, package_name)
                    extends_argument_classes.append(extends_argument_package_class)
                extends_package_class_list += extends_argument_classes
                return extends_package_class + '<' + ','.join(extends_argument_classes) + '>', extends_package_class_list
            else:
                return extends_package_class, [extends_package_class]
        else:
            extends_class = self._get_extends_class_full_package(extends.name, import_list, package_name)
            return extends_class, [extends_class]

    def _get_extends_class_full_package(self, extends_class, import_list, package_name):
        extends_in_imports = [import_obj for import_obj in import_list if extends_class in import_obj['import_path']]
        return extends_in_imports[0]['import_path'] if extends_in_imports else package_name + '.' + extends_class

    def _parse_class(self, node, filepath: str, package_name: str, import_list: list, commit_or_branch: str):
        # 处理class信息
        documentation = node.documentation
        class_name = node.name
        package_class = package_name + '.' + node.name
        class_type = type(node).__name__.replace('Declaration', '')
        access_modifier = [m for m in list(node.modifiers) if m.startswith('p')][0] if list([m for m in list(node.modifiers) if m.startswith('p')]) else 'public'
        annotations_json = json.dumps(node.annotations, default=lambda obj: obj.__dict__)
        is_controller, controller_base_url = self._judge_is_controller(node.annotations)
        extends_package_class = None
        if 'extends' in node.attrs and node.extends:
            extends_package_class, extends_package_class_list = self._handle_extends(node.extends, import_list, package_name)
            package_path = package_class.replace('.', '/') + '.java'
            base_filepath = filepath.replace(package_path, '')
            for extends_package_class_item in extends_package_class_list:
                extends_package = '.'.join(extends_package_class_item.split('.')[0: -1])
                extends_class_name = extends_package_class_item.split('.')[-1]
                extends_class_filepath = base_filepath + extends_package_class_item.replace('.', '/') + '.java'
                extends_class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE project_id={self.project_id} and package_name = "{extends_package}" and class_name = "{extends_class_name}" and filepath= "{extends_class_filepath}"')
                # if not extends_class_db:
                # self.parse_java_file(extends_class_filepath, commit_or_branch)
        implements = ','.join([implement.name for implement in node.implements]) if 'implements' in node.attrs and node.implements else None
        class_id, new_add = self.sqlite.add_class(filepath.replace('\\', '/'), access_modifier, class_type, class_name, package_name, extends_package_class, self.project_id, implements, annotations_json, documentation, is_controller, controller_base_url, commit_or_branch)
        return class_id, new_add

    def _parse_imports(self, imports):
        import_list = []
        for import_decl in imports:
            import_obj = {
                'import_path': import_decl.path,
                'is_static': import_decl.static,
                'is_wildcard': import_decl.wildcard,
                'start_line': import_decl.position.line,
                'end_line': import_decl.position.line
            }
            import_list.append(import_obj)
        return import_list

    def _parse_fields(self, fields, package_name, class_id, import_map):
        field_list = []
        for field_obj in fields:
            field_annotations = json.dumps(field_obj.annotations, default=lambda obj: obj.__dict__)
            access_modifier = [m for m in list(field_obj.modifiers) if m.startswith('p')][0] if list([m for m in list(field_obj.modifiers) if m.startswith('p')]) else 'public'
            field_name = field_obj.declarators[0].name
            field_type: str = field_obj.type.name
            if field_type.lower() not in JAVA_BASIC_TYPE:
                if field_type in import_map.keys():
                    field_type = import_map.get(field_type)
                else:
                    import_map[field_obj.type.name] = package_name + '.' + field_obj.type.name
                    field_type = package_name + '.' + field_type
            is_static = 'static' in list(field_obj.modifiers)
            documentation = field_obj.documentation
            start_line = field_obj.position.line if not field_obj.annotations else field_obj.annotations[0].position.line
            end_line = field_obj.position.line
            field_obj = {
                'class_id': class_id,
                'project_id': self.project_id,
                'annotations': field_annotations,
                'access_modifier': access_modifier,
                'field_type': field_type,
                'field_name': field_name,
                'is_static': is_static,
                'documentation': documentation,
                'start_line': start_line,
                'end_line': end_line
            }
            field_list.append(field_obj)
        self.sqlite.update_data(f'DELETE FROM field where class_id={class_id}')
        self.sqlite.insert_data('field', field_list)
        return field_list

    def _parse_method(self, methods, lines, class_id, import_map, field_map):
        # 处理 methods
        all_method = []
        class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE class_id = {class_id}')[0]
        base_url = class_db['controller_base_url'] if class_db['controller_base_url'] else ''
        method_name_entity_map = {method.name: method for method in methods}
        for method_obj in methods:
            method_invocation = {}
            variable_map = {}
            method_name = method_obj.name
            documentation = method_obj.documentation  # document
            annotations = json.dumps(method_obj.annotations, default=lambda obj: obj.__dict__)  # annotations
            is_api, api_path = self._judge_is_api(method_obj.annotations, base_url, method_name)
            access_modifier = [m for m in list(method_obj.modifiers) if m.startswith('p')][0] if list([m for m in list(method_obj.modifiers) if m.startswith('p')]) else 'public'
            is_static = 'static' in list(method_obj.modifiers)
            is_abstract = 'abstract' in list(method_obj.modifiers)
            # 处理返回对象
            return_type = self._deal_declarator_type(method_obj.return_type, import_map, method_invocation, RETURN_TYPE)
            method_start_line = method_obj.position.line
            if method_obj.annotations:
                method_start_line = method_obj.annotations[0].position.line
            method_end_line = self._get_method_end_line(method_obj)
            method_body = lines[method_start_line - 1: method_end_line + 1]
            # 处理参数
            parameters = []
            for parameter in method_obj.parameters:
                parameter_obj = {
                    'parameter_type': self._deal_declarator_type(parameter.type, import_map, method_invocation, PARAMETERS),
                    'parameter_name': parameter.name,
                    'parameter_varargs': parameter.varargs
                }
                parameters.append(parameter_obj)
            parameters_map = {parameter['parameter_name']: parameter['parameter_type'] for parameter in parameters}

            # 处理方法体
            if method_obj.body:
                for body in method_obj.body:
                    for path, node in body.filter(javalang.tree.VariableDeclaration):
                        var_declarator = node.declarators[0].name
                        var_declarator_type = self._deal_declarator_type(node.type, import_map, method_invocation, BODY)
                        variable_map[var_declarator] = var_declarator_type
                    for path, node in body.filter(javalang.tree.MethodInvocation):
                        qualifier = node.qualifier
                        member = node.member
                        # 类静态方法调用
                        if not qualifier and not member[0].islower():
                            qualifier_type = self._get_var_type(member, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                            # todo a.b.c
                            if node.selectors is None:
                                continue
                            for selector in node.selectors:
                                selector_member = selector.member
                                selector_arguments = self._deal_var_type(selector.arguments, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                                selector_line = selector.position.line
                                selector_method = f'{selector_member}({",".join(selector_arguments)})'
                                self._add_method_used_to_method_invocation(method_invocation, qualifier_type, selector_method, [selector_line])
                        elif qualifier:
                            qualifier_type = self._get_var_type(qualifier, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                            node_arguments = self._deal_var_type(node.arguments, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                            node_line = node.position.line
                            node_method = f'{member}({",".join(node_arguments)})'
                            self._add_method_used_to_method_invocation(method_invocation, qualifier_type, node_method, [node_line])
                        # 在一个类的方法或父类方法
                        elif member:
                            class_db = self.sqlite.select_data(f'SELECT * FROM class where class_id={class_id} limit 1')[0]
                            package_class = class_db['package_name'] + '.' + class_db['class_name']
                            node_line = node.position.line
                            node_arguments = self._deal_var_type(node.arguments, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                            # todo 同级方法, 判断参数长度，不精确
                            if method_name_entity_map.get(member):
                                same_class_method = None
                                max_score = -float('inf')
                                for method_item in methods:
                                    if method_item.name != member or len(node.arguments) != len(method_item.parameters):
                                        continue
                                    method_item_param_types = [self._deal_declarator_type(parameter.type, import_map, method_invocation, PARAMETERS) for parameter in method_item.parameters]
                                    score = self._calculate_similar_score_method_params(node_arguments, method_item_param_types)
                                    if score > max_score:
                                        max_score = score
                                        same_class_method = method_item
                                if same_class_method:
                                    node_arguments = self._deal_var_type(same_class_method.parameters, parameters_map, variable_map, field_map, import_map, method_invocation, BODY)
                                    node_method = f'{member}({",".join(node_arguments)})'
                                    self._add_method_used_to_method_invocation(method_invocation, package_class, node_method, [node_line])
                            # todo 继承方法
                            elif class_db['extends_class']:
                                extends_package_class, method_params = self._find_method_in_extends(class_db['extends_class'], member, node_arguments)
                                if not extends_package_class:
                                    self._add_method_used_to_method_invocation(method_invocation, extends_package_class, method_params, [node_line])

                    # for path, node in body.filter(javalang.tree.SuperMethodInvocation):
                    #     print(node)
                    # for path, node in body.filter(javalang.tree.TypeArgument):
                    #     print(node)
                    # for path, node in body.filter(javalang.tree.TypeParameter):
                    #     print(node)
                    # for path, node in body.filter(javalang.tree.FieldDeclaration):
                    #     print(node)
            method_db = {
                'class_id': class_id,
                'project_id': self.project_id,
                'annotations': annotations,
                'access_modifier': access_modifier,
                'return_type': return_type,
                'method_name': method_name,
                'parameters': json.dumps(parameters),
                'body': json.dumps(method_body),
                'method_invocation_map': json.dumps(method_invocation),
                'is_static': is_static,
                'is_abstract': is_abstract,
                'is_api': is_api,
                'api_path': json.dumps(api_path),
                'start_line': method_start_line,
                'end_line': method_end_line,
                'documentation': documentation
            }
            all_method.append(method_db)
        self.sqlite.update_data(f'DELETE FROM methods where class_id={class_id}')
        self.sqlite.insert_data('methods', all_method)

    def _find_method_in_extends(self, extend_package_class: str, method_name: str, method_arguments):
        if not extend_package_class:
            return None, None
        # 查表有没有记录
        extend_package = '.'.join(extend_package_class.split('.')[0: -1])
        extend_class = extend_package_class.split('.')[-1]
        extend_class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE package_name="{extend_package}" '
                                                  f'AND class_name="{extend_class}" '
                                                  f'AND project_id={self.project_id} limit 1')

        if not extend_class_db:
            return None, None
        extend_class_entity = extend_class_db[0]
        extend_class_id = extend_class_entity['class_id']
        methods_db_list = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id={extend_class_id} and method_name = "{method_name}"')
        if not methods_db_list and not extend_class_entity['extends_class']:
            return None, None
        if not methods_db_list:
            return self._find_method_in_extends(extend_class_entity['extends_class'], method_name, method_arguments)
        else:
            filter_methods = [method for method in methods_db_list if len(json.loads(method['parameters'])) == len(method_arguments)]
            if not filter_methods:
                return self._find_method_in_extends(extend_class_entity['extends_class'], method_name, method_arguments)
            package_class = extend_class_entity['package_name'] + '.' + extend_class_entity['class_name']
            if len(filter_methods) == 1:
                method_db = filter_methods[0]
                method_params = f'{method_db["method_name"]}({",".join([param["parameter_type"] for param in json.loads(method_db["parameters"])])})'
                return package_class, method_params
            else:
                max_score = -float('inf')
                max_score_method = None
                for method_db in filter_methods:
                    method_db_params = [param["parameter_type"] for param in json.loads(method_db["parameters"])]
                    score = self._calculate_similar_score_method_params(method_arguments, method_db_params)
                    if score > max_score:
                        max_score = score
                        max_score_method = method_db
                if max_score_method is None:
                    max_score_method = filter_methods[0]
                method_params = f'{max_score_method["method_name"]}({",".join([param["parameter_type"] for param in json.loads(max_score_method["parameters"])])})'
                return package_class, method_params

    def _calculate_similar_score_method_params(self, except_method_param_list, method_param_list):
        score = 0
        positions = {}

        # 记录list1中每个元素的位置
        for i, item in enumerate(except_method_param_list):
            positions[item] = i

        # 遍历list2,计算分数
        for i, item in enumerate(method_param_list):
            if item in positions:
                score += 1
                score -= abs(i - positions[item])

        return score

    def _get_method_end_line(self, method_obj):
        method_end_line = method_obj.position.line
        while True:
            if isinstance(method_obj, list):
                if None in method_obj:
                    method_obj.remove(None)
                if len(method_obj) == 0:
                    break
                length = len(method_obj)
                for i in range(0, length):
                    temp = method_obj[length - 1 - i]
                    if temp is not None:
                        method_obj = temp
                        break
                if method_obj is None:
                    break
            if isinstance(method_obj, list):
                continue
            if hasattr(method_obj, 'position') \
                    and method_obj.position is not None \
                    and method_obj.position.line > method_end_line:
                method_end_line = method_obj.position.line
            if hasattr(method_obj, 'children'):
                method_obj = method_obj.children
            else:
                break
        return method_end_line

    def _get_element_value(self, method_element):
        method_api_path = []
        if type(method_element).__name__ == 'BinaryOperation':
            operandl = method_element.operandl
            operandr = method_element.operandr
            operandl_str = self._get_api_part_route(operandl)
            operandr_str = self._get_api_part_route(operandr)
            method_api_path = [operandl_str + operandr_str]
        elif type(method_element).__name__ == 'MemberReference':
            method_api_path = [method_element.member.replace('"', '')]
        elif type(method_element).__name__ == 'ElementArrayValue':
            method_api_path = self._get_element_with_values(method_element)
        elif method_element.value is not None:
            method_api_path = [method_element.value.replace('"', '')]
        return method_api_path

    def _get_element_with_values(self, method_api_path_obj):
        result = []
        for method_api_value in method_api_path_obj.values:
            result += self._get_element_value(method_api_value)
        return result

    def _get_api_part_route(self, part):
        part_class = type(part).__name__
        if part_class == 'MemberReference':
            return part.member.replace('"', '')
        elif part_class == 'Literal':
            return part.value.replace('"', '')

    def _judge_is_controller(self, annotation_list):
        is_controller = any('Controller' in annotation.name for annotation in annotation_list)
        base_request = ''
        if not is_controller:
            return is_controller, base_request
        for annotation in annotation_list:
            if 'RequestMapping' != annotation.name:
                continue
            if annotation.element is None:
                continue
            if isinstance(annotation.element, list):
                base_request_list = []
                for annotation_element in annotation.element:
                    if annotation_element.name != 'value' and annotation_element.name != 'path':
                        continue
                    if 'values' in annotation_element.value.attrs:
                        base_request_list += self._get_element_with_values(annotation_element.value)
                    else:
                        base_request_list += self._get_element_value(annotation_element.value)
                if len(base_request_list) > 0:
                    base_request = base_request_list[0]
            else:
                if 'value' in annotation.element.attrs:
                    base_request = annotation.element.value.replace('"', '')
                elif 'values' in annotation.element.attrs:
                    base_request = ' || '.join([literal.value for literal in annotation.element.values])
        if is_controller and not base_request.endswith('/'):
            base_request += '/'
        return is_controller, base_request

    def _judge_is_api(self, method_annotations, base_request, method_name):
        api_path_list = []
        req_method_list = []
        method_api_path = []
        is_api = False
        for method_annotation in method_annotations:
            if 'Mapping' not in method_annotation.name:
                continue
            is_api = True
            if method_annotation.name != 'RequestMapping':
                req_method_list.append(method_annotation.name.replace('Mapping', ''))
            else:
                for method_annotation_element in method_annotation.element:
                    if type(method_annotation_element) == tuple:
                        req_method_list = ['ALL']
                        break
                    if 'name' not in method_annotation_element.attrs or method_annotation_element.name != 'method':
                        continue
                    method_annotation_element_value = method_annotation_element.value
                    if 'member' in method_annotation_element_value.attrs:
                        req_method_list.append(method_annotation_element_value.member)
                    elif 'values' in method_annotation_element_value.attrs:
                        method_annotation_element_values = method_annotation_element_value.values
                        req_method_list += [method_annotation_element_temp.member for
                                            method_annotation_element_temp in
                                            method_annotation_element_values
                                            if 'member' in method_annotation_element_temp.attrs]
            if not isinstance(method_annotation.element, list):
                if method_annotation.element is None:
                    continue
                method_api_path += self._get_element_value(method_annotation.element)
            else:
                method_api_path_list = [method_annotation_element.value for method_annotation_element in method_annotation.element
                                        if method_annotation_element.name == 'path' or method_annotation_element.name == 'value']
                if len(method_api_path_list) == 0:
                    continue
                method_api_path_obj = method_api_path_list[0]
                if 'value' in method_api_path_obj.attrs:
                    method_api_path += [method_api_path_obj.value.replace('"', '')]
                else:
                    if 'values' in method_api_path_obj.attrs:
                        for method_api_value in method_api_path_obj.values:
                            method_api_path += self._get_element_value(method_api_value)
                    else:
                        method_api_path += [method_name + '/cci-unknown']
        if len(method_api_path) == 0:
            method_api_path = ['/']
        for method_api_path_obj in method_api_path:
            if method_api_path_obj.startswith('/'):
                method_api_path_obj = method_api_path_obj[1:]
            api_path = base_request + method_api_path_obj
            if api_path.endswith('/'):
                api_path = api_path[0:-1]
            if len(req_method_list) > 0:
                api_path_list += ['[' + req_method_temp + ']' + api_path for req_method_temp in req_method_list]
            else:
                api_path_list.append('[ALL]' + api_path)
        return is_api, api_path_list

    def _add_entity_used_to_method_invocation(self, method_invocation, package_class, section):
        if package_class not in method_invocation.keys():
            method_invocation[package_class] = {ENTITY: {section: True}}
        elif ENTITY not in method_invocation[package_class].keys():
            method_invocation[package_class][ENTITY] = {section: True}
        elif section not in method_invocation[package_class][ENTITY].keys():
            method_invocation[package_class][ENTITY][section] = True

    def _add_method_used_to_method_invocation(self, method_invocation, package_class, method, lines):
        if package_class not in method_invocation.keys():
            method_invocation[package_class] = {METHODS: {method: lines}}
        elif METHODS not in method_invocation[package_class].keys():
            method_invocation[package_class][METHODS] = {method: lines}
        elif method not in method_invocation[package_class][METHODS].keys():
            method_invocation[package_class][METHODS][method] = lines
        else:
            method_invocation[package_class][METHODS][method] += lines

    def _add_field_used_to_method_invocation(self, method_invocation, package_class, field, lines):
        if package_class not in method_invocation.keys():
            method_invocation[package_class] = {FIELDS: {field: lines}}
        elif FIELDS not in method_invocation[package_class].keys():
            method_invocation[package_class][FIELDS] = {field: lines}
        elif field not in method_invocation[package_class][FIELDS].keys():
            method_invocation[package_class][FIELDS][field] = lines
        else:
            method_invocation[package_class][FIELDS][field] += lines

    def _deal_declarator_type(self, node_type, import_map, method_invocation, section):
        if node_type is None:
            return node_type
        if type(node_type) == javalang.tree.BasicType:
            return node_type.name
        var_declarator_type = node_type.name
        if var_declarator_type in import_map.keys():
            var_declarator_type = import_map.get(var_declarator_type)
            self._add_entity_used_to_method_invocation(method_invocation, var_declarator_type, section)
        var_declarator_type_arguments = self._deal_arguments_type(node_type.arguments, import_map, method_invocation, section)
        if var_declarator_type_arguments:
            var_declarator_type = var_declarator_type + '<' + '#'.join(var_declarator_type_arguments) + '>'
        return var_declarator_type

    def _deal_arguments_type(self, arguments, import_map, method_invocation, section):
        var_declarator_type_arguments_new = []
        if not arguments:
            return var_declarator_type_arguments_new
        var_declarator_type_arguments = []
        for argument in arguments:
            if type(argument) == javalang.tree.MethodInvocation:
                var_declarator_type_arguments.append(PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN)
                continue
            var_declarator_type_argument = self._deal_type(argument)
            if var_declarator_type_argument in import_map.keys():
                var_declarator_type_argument = import_map.get(var_declarator_type_argument)
                self._add_entity_used_to_method_invocation(method_invocation, var_declarator_type_argument, section)
            var_declarator_type_arguments.append(var_declarator_type_argument)
        return var_declarator_type_arguments

    def _deal_type(self, argument):
        if not argument:
            return None
        argument_type = type(argument)
        if argument_type == javalang.tree.MemberReference:
            var_declarator_type_argument = argument.member
        elif argument_type == javalang.tree.Literal:
            var_declarator_type_argument = self._deal_literal_type(argument.value)
        elif argument_type == javalang.tree.This:
            var_declarator_type_argument = 'This'
        elif argument_type == javalang.tree.LambdaExpression:
            var_declarator_type_argument = 'LambdaExpression'
        elif argument_type == javalang.tree.BinaryOperation:
            # todo BinaryOperation temp set string
            var_declarator_type_argument = 'String'
        elif argument_type == javalang.tree.MethodReference or argument_type == javalang.tree.TernaryExpression:
            # todo MethodReference temp set unknown
            var_declarator_type_argument = PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
        elif argument_type == javalang.tree.SuperMethodInvocation:
            logging.info(argument_type)
            var_declarator_type_argument = PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
        elif argument_type == javalang.tree.Assignment:
            var_declarator_type_argument = self._deal_type(argument.value)
        elif argument_type == javalang.tree.Cast:
            var_declarator_type_argument = argument.type.name
        # todo
        elif argument_type == javalang.tree.SuperMemberReference:
            var_declarator_type_argument = 'String'
        elif argument.type is not None:
            var_declarator_type_argument = argument.type.name
        else:
            logging.info(f'argument type is None：{argument}')
            var_declarator_type_argument = PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
        return var_declarator_type_argument

    def _deal_literal_type(self, text):
        if 'true' == text or 'false' == text:
            return 'boolean'
        if text.isdigit():
            return 'int'
        return 'String'

    def _deal_var_type(self, arguments, parameters_map, variable_map, field_map, import_map, method_invocation, section):
        var_declarator_type_arguments_new = []
        if not arguments:
            return var_declarator_type_arguments_new
        var_declarator_type_arguments = []
        for argument in arguments:
            argument_type = type(argument)
            if argument_type == javalang.tree.MethodInvocation:
                var_declarator_type_arguments.append(PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN)
                continue
            var_declarator_type_argument = self._deal_type(argument)
            var_declarator_type_argument = self._get_var_type(var_declarator_type_argument, parameters_map, variable_map, field_map, import_map, method_invocation, section)
            type_arguments = self._deal_arguments_type(argument.type.arguments, import_map, method_invocation, section) \
                if 'type' in argument.attrs \
                   and not isinstance(argument.type, str) \
                   and 'arguments' in argument.type.attrs \
                   and argument.type.arguments \
                else []
            if type_arguments:
                var_declarator_type_argument = var_declarator_type_argument + '<' + '#'.join(type_arguments) + '>'
            var_declarator_type_arguments.append(var_declarator_type_argument)
        return var_declarator_type_arguments

    def _get_var_type(self, var, parameters_map, variable_map, field_map, import_map, method_invocation, section):
        if not var:
            return var
        if var in parameters_map.keys():
            return parameters_map.get(var)
        if var in variable_map.keys():
            return variable_map.get(var)
        if var in field_map.keys():
            field_type = field_map.get(var)['field_type']
            package_class = field_map.get(var)['package_class']
            start_line = field_map.get(var)['start_line']
            self._add_field_used_to_method_invocation(method_invocation, package_class, var, [start_line])
            return field_type
        if var in import_map.keys():
            var_type = import_map.get(var)
            self._add_entity_used_to_method_invocation(method_invocation, var_type, section)
            return var_type
        return var

    def _get_extends_class_fields_map(self, class_id: int):
        class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE class_id = {class_id}')[0]
        extend_package_class = class_db['extends_class']
        if not extend_package_class:
            return {}
        extend_package = '.'.join(extend_package_class.split('.')[0: -1])
        extend_class = extend_package_class.split('.')[-1]
        extend_class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE package_name="{extend_package}" '
                                                  f'AND class_name="{extend_class}" '
                                                  f'AND project_id={self.project_id} limit 1')
        if not extend_class_db:
            return {}
        extend_class_entity = extend_class_db[0]
        extend_class_id = extend_class_entity['class_id']
        extend_class_fields = self.sqlite.select_data(f'SELECT * FROM field WHERE class_id = {extend_class_id}')
        extend_class_fields_map = {field_obj['field_name']: {'field_type': field_obj['field_type'], 'package_class': extend_package_class, 'start_line': field_obj['start_line']} for field_obj in extend_class_fields}
        if not extend_class_entity['extends_class']:
            return extend_class_fields_map
        else:
            extend_new_map = self._get_extends_class_fields_map(extend_class_id)
            extend_new_map.update(extend_class_fields_map)
            return extend_new_map

    def parse_java_file(self, filepath, commit_or_branch):
        if not filepath.endswith('.java'):
            return
        try:
            with open(filepath, encoding='UTF-8') as fp:
                file_content = fp.read()
        except:
            return
        logging.info(f'Parsing java file: {filepath}')
        lines = file_content.split('\n')
        try:
            tree = javalang.parse.parse(file_content)
            if not tree.types:
                return
        except Exception as e:
            logging.error(f"Error parsing {filepath}: {e}")
            return
        # 处理包信息
        package_name = tree.package.name if tree.package else None
        class_name = tree.types[0].name
        package_class = package_name + '.' + tree.types[0].name
        # 处理 import 信息
        import_list = self._parse_imports(tree.imports)
        import_map = {import_obj['import_path'].split('.')[-1]: import_obj['import_path'] for import_obj in import_list}

        # 处理 class 信息
        class_id, new_add = self._parse_class(tree.types[0], filepath, package_name, import_list, commit_or_branch)
        # 已经处理过了，返回
        # if not new_add:
        #     return
        # 导入import
        imports = [dict(import_obj, class_id=class_id, project_id=self.project_id) for import_obj in import_list]
        self.sqlite.update_data(f'DELETE FROM import WHERE class_id={class_id}')
        self.sqlite.insert_data('import', imports)

        # 处理 field 信息
        field_list = self._parse_fields(tree.types[0].fields, package_name, class_id, import_map)
        field_map = {field_obj['field_name']: {'field_type': field_obj['field_type'], 'package_class': package_class, 'start_line': field_obj['start_line']} for field_obj in field_list}
        import_map = dict((k, v) for k, v in import_map.items() if v.startswith('com.patsnap'))

        # 将extend class的field导进来
        extends_class_fields_map = self._get_extends_class_fields_map(class_id)
        extends_class_fields_map.update(field_map)
        # 处理 methods 信息
        self._parse_method(tree.types[0].methods, lines, class_id, import_map, extends_class_fields_map)

    def parse_java_file_list(self, filepath_list: list, commit_or_branch: str):
        for file_path in filepath_list:
            self.parse_java_file(file_path, commit_or_branch)


if __name__ == '__main__':
    print('jcci')



from src.jcci import jcci

#
# jcci.analyze('git@git.patsnap.com:platform/project-identity.git',
#              'master',
#              '92bdd486',
#              '3887d5c0', 'username1')
# # service and xnl
# jcci.analyze('git@git.patsnap.com:core-product/qa/libs/s-autotest-api-admin.git',
#              'master',
#              '9770b53',
#              'f4a166e', 'username1')

# jcci.analyze('git@git.patsnap.com:core-product/qa/libs/s-autotest-api-admin.git',
#              'master',
#              '73ca7b62e60d79eba018c5c8e4be90',
#              'b90b7387b735ba784a7cf316e8e153', 'username1')
#
# jcci.analyze('git@git.patsnap.com:core-product/qa/libs/s-autotest-api-admin.git',
#              'master',
#              '016f94b84b64',
#              '5c893f874892', 'username1')

# jcci.analyze('git@git.patsnap.com:core-product/backend/s-analytics-workspace.git',
#              'master',
#              'a0954b079cf39db14cd0c073ecd1d9146341b27d',
#              'bb779398cc45c2864a3566eed629d8806013d18e',
#              'username1')

# extends
jcci.analyze('git@git.patsnap.com:core-product/backend/s-analytics-workspace.git',
             'master',
             '56a700827e658064f41630145694e255432c2caf',
             '6a6e39c2d9a6a1e6e24183e96186dbdce624ab01',
             'username1')

# jcci.analyze('git@git.patsnap.com:core-product/backend/s-core-eureka.git',
#              'master',
#              '60562f1e',
#              '8f71adf7', 'username1')
#
# jcci.analyze('git@git.patsnap.com:platform/project-backoffice.git',
#              'master',
#              '090d7b7b',
#              'e9afe2ca', 'username1')
#



# jcci.analyze('git@git.patsnap.com:core-product/qa/libs/s-autotest-api-admin.git',
#              'master',
#              'c5c093d',
#              '7334f8e', 'username1')

#
# jcci.analyze('git@git.patsnap.com:platform/project-user-service.git',
#              'master',
#              '681c804',
#              'e18a5ea', 'username1')

#
# jcci.analyze('git@gitee.com:chen-shaoxing/code-chain-demo.git',
#              'master',
#              'f82358d',
#              'fd3adfb', 'username1')

#
# jcci.analyze_branches('git@git.patsnap.com:core-product/qa/libs/s-autotest-api-admin.git',
#              'master',
#              'release/test2','username1')


#
#
# jcci.analyze('https://gitee.com/youlaiorg/youlai-cloud.git',
#              'master',
#              '5e66c11','1146b37','username1')




create_tables = '''
CREATE TABLE project (
	project_id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	project_name TEXT NOT NULL,
    git_url TEXT NOT NULL,
    branch TEXT NOT NULL,
    commit_or_branch_new TEXT NOT NULL,
    commit_or_branch_old TEXT,
    create_at TIMESTAMP NOT NULL DEFAULT (datetime('now','localtime'))
);

CREATE TABLE class (
	class_id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	filepath TEXT,
	access_modifier TEXT,
	class_type TEXT NOT NULL,
	class_name TEXT NOT NULL,
	package_name TEXT NOT NULL,
	extends_class TEXT,
	project_id INTEGER NOT NULL,
	implements TEXT,
	annotations TEXT,
	documentation TEXT,
	is_controller REAL,
	controller_base_url TEXT,
	commit_or_branch TEXT,
	create_at TIMESTAMP NOT NULL DEFAULT (datetime('now','localtime'))
);

CREATE TABLE import (
	import_id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	class_id INTEGER NOT NULL,
	project_id INTEGER NOT NULL,
	start_line INTEGER,
	end_line INTEGER,
	import_path TEXT,
	is_static REAL,
	is_wildcard REAL,
    create_at TIMESTAMP NOT NULL DEFAULT (datetime('now','localtime'))
);

CREATE TABLE field (
	field_id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	class_id INTEGER,
	project_id INTEGER NOT NULL,
	annotations TEXT,
	access_modifier TEXT,
	field_type TEXT,
	field_name TEXT,
	is_static REAL,
	start_line INTEGER,
	end_line INTEGER,
	documentation TEXT,
    create_at TIMESTAMP NOT NULL DEFAULT (datetime('now','localtime'))
);


CREATE TABLE methods (
	method_id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	class_id INTEGER NOT NULL,
	project_id INTEGER NOT NULL,
	annotations TEXT,
	access_modifier TEXT,
	return_type TEXT,
	method_name TEXT NOT NULL,
	parameters TEXT,
	body TEXT,
	method_invocation_map TEXT,
	is_static REAL,
	is_abstract REAL,
	is_api REAL,
	api_path TEXT,
	start_line INTEGER NOT NULL,
	end_line INTEGER NOT NULL,
	documentation TEXT,
    create_at TIMESTAMP NOT NULL DEFAULT (datetime('now','localtime'))
);
'''




# -*- coding: UTF-8 -*-
import os

db_path = os.path.dirname(os.path.abspath(__file__))
project_path = os.path.dirname(os.path.abspath(__file__))
ignore_file = ['*/pom.xml', '*/test/*', '*.sh', '*.md', '*/checkstyle.xml', '*.yml', '.git/*']




import unidiff
import os


def _diff_patch_lines(patch):
    line_num_added = []
    line_num_removed = []
    line_content_added = []
    line_content_removed = []
    for hunk in patch:
        if hunk.added > 0:
            targets = hunk.target
            target_start = hunk.target_start
            for i in range(0, len(targets)):
                if targets[i].startswith('+') and not targets[i][1:].strip().startswith(('*', '//', 'import ')) and targets[i][1:].strip():
                    line_num_added.append(target_start + i)
                    line_content_added.append(targets[i][1:])
        if hunk.removed > 0:
            sources = hunk.source
            source_start = hunk.source_start
            for i in range(0, len(sources)):
                if sources[i].startswith('-') and not sources[i][1:].strip().startswith(('*', '//', 'import ')) and sources[i][1:].strip():
                    line_num_removed.append(source_start + i)
                    line_content_removed.append(sources[i][1:])
    return line_num_added, line_content_added, line_num_removed, line_content_removed


def get_diff_info(file_path):
    patch_results = {}
    with open(file_path, encoding='UTF-8') as f:
        diff_text = f.read()
    patch_set = unidiff.PatchSet(diff_text)
    for patch in patch_set:
        if '.git' in patch.path or os.path.join('src', 'test') in patch.path or (not patch.path.endswith(('.java', '.xml'))):
            continue
        line_num_added, line_content_added, line_num_removed, line_content_removed = _diff_patch_lines(patch)
        java_file_path = patch.path
        patch_results[java_file_path] = {
            'line_num_added': line_num_added,
            'line_content_added': line_content_added,
            'line_num_removed': line_num_removed,
            'line_content_removed': line_content_removed
        }
    return patch_results


if __name__ == '__main__':
    print('jcci')




import json
import os
import sys
import time
import atexit
import logging
import datetime
import fnmatch
from src.jcci.config import project_path, db_path, ignore_file
from src.jcci.database import SqliteHelper
from src.jcci.java_parse import JavaParse
import src.jcci.mapper_parse as mapper_parse
import src.jcci.diff_parse as diff_parse
import src.jcci.graph as graph
import src.jcci.constant as constant
from itertools import combinations

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.DEBUG)


class JCCI(object):
    def __init__(self, git_url, username):
        self.git_url = git_url
        self.username: str = username
        self.branch_name: str = ''
        self.commit_or_branch_new: str = ''
        self.commit_or_branch_old: str = ''
        self.project_id: int = -1
        self.cci_filepath: str = ''
        self.project_name: str = ''
        self.file_path: str = ''
        self.sqlite = SqliteHelper(db_path + '/' + username + '_jcci.db')
        self.view = graph.Graph()
        self.t1 = datetime.datetime.now()
        self.need_analyze_obj_list = []
        self.analyzed_obj_set = []
        self.diff_parse_map = {}
        self.xml_parse_results_new = {}
        self.xml_parse_results_old = {}

    # Step 1.1
    def _can_analyze(self, filepath, cci_file_path):
        # 已有分析结果
        if os.path.exists(cci_file_path):
            logging.info('Has analyze result, skip!')
            with open(cci_file_path, 'r') as read:
                result = read.read()
                result_json = json.loads(result)
                print(result, flush=True)
                print(f'Impacted api list: {result_json["impacted_api_list"]}', flush=True)
            sys.exit(0)

        # 正在分析
        wait_index = 0
        occupy_filepath = os.path.join(filepath, 'Occupy.ing')
        atexit.register(self._clean_occupy, occupy_filepath)
        while os.path.exists(occupy_filepath) and wait_index < 30:
            logging.info(f'Analyzing by others, waiting or clean occupying file manually at: {occupy_filepath} to continue')
            time.sleep(3)
            wait_index += 1
        if os.path.exists(occupy_filepath):
            logging.info(f'Analyzing by others, waiting timeout')
            sys.exit(0)

    # Step 1.2
    def _clean_occupy(self, occupy_path):
        if os.path.exists(occupy_path):
            os.remove(occupy_path)

    # Step 1.3
    def _occupy_project(self):
        # 占住项目分析
        logging.info('Start occupying project, and others can not analyze until released')
        occupy_filepath = os.path.join(self.file_path, 'Occupy.ing')
        with open(occupy_filepath, 'w') as ow:
            ow.write(f'Occupy by {self.username}')
        time.sleep(1)

    # Step 2
    def _get_diff_parse_map(self, filepath, branch, commit_first, commit_second):
        logging.info('Git pull project to HEAD')
        os.system(f'cd {filepath} && git checkout {branch} && git pull')
        time.sleep(1)
        logging.info(f'Git diff between {commit_first} and {commit_second}')
        diff_base = f'cd {self.file_path} && git diff {commit_second}..{commit_first} > diff_{commit_second}..{commit_first}.txt'
        os.system(diff_base)
        diff_txt = os.path.join(self.file_path, f'diff_{commit_second}..{commit_first}.txt')
        logging.info(f'Analyzing diff file, location: {diff_txt}')
        return diff_parse.get_diff_info(diff_txt)

    # Step 3
    def _parse_project(self, project_dir, new_commit_or_branch, old_commit_or_branch):
        # 解析最新的项目文件
        os.system(f'cd {project_dir} && git reset --hard {new_commit_or_branch}')
        time.sleep(2)
        file_path_list = self._get_project_files(project_dir)
        diff_xml_file_path = [key for key in file_path_list if key.endswith('.xml') and any(key.endswith(diff_path) for diff_path in self.diff_parse_map.keys())]
        java_parse = JavaParse(self.sqlite.db_path, self.project_id)
        java_parse.parse_java_file_list(file_path_list, new_commit_or_branch)
        xml_parse_result_new = self._parse_xml_file(diff_xml_file_path)
        xml_parse_result_old = {}
        if not old_commit_or_branch:
            return xml_parse_result_new, xml_parse_result_old
        # 解析旧版本有差异的文件
        os.system(f'cd {project_dir} && git reset --hard {old_commit_or_branch}')
        time.sleep(2)
        xml_parse_result_old = self._parse_xml_file(diff_xml_file_path)
        for key in self.diff_parse_map.keys():
            matched_file_path_list = [filepath for filepath in file_path_list if filepath.endswith(key)]
            if not matched_file_path_list:
                continue
            matched_file_path = matched_file_path_list[0]
            java_parse.parse_java_file(matched_file_path, old_commit_or_branch)
        return xml_parse_result_new, xml_parse_result_old

    # Step 3.1 get all java files
    def _get_project_files(self, project_dir):
        file_lists = []
        for root, dirs, files in os.walk(project_dir):
            if '.git' in root or os.path.join('src', 'test') in root:
                continue
            for file in files:
                ignore = False
                filepath = os.path.join(root, file)
                for pattern in ignore_file:
                    if fnmatch.fnmatch(filepath, pattern):
                        ignore = True
                        break
                if ignore:
                    continue
                filepath.replace('\\', '/')
                file_lists.append(filepath)
        return file_lists

    # Step 3.3
    def _parse_xml_file(self, file_path_list):
        xml_parse_results = {}
        for filepath in file_path_list:
            if filepath.endswith('.xml'):
                xml_parse_result = mapper_parse.parse(filepath)
                xml_parse_results[filepath] = xml_parse_result
        return xml_parse_results

    # Step 4
    def _diff_analyze(self, patch_filepath: str, diff_parse_obj: dict):
        is_xml_file = patch_filepath.endswith('.xml')
        if is_xml_file:
            self._xml_diff_analyze(patch_filepath, diff_parse_obj)
        else:
            self._java_diff_analyze(patch_filepath, diff_parse_obj)

    # Step 4.1
    def _xml_diff_analyze(self, patch_filepath, diff_parse_obj: dict):
        xml_file_path_list = [filepath for filepath in self.xml_parse_results_new.keys() if self.file_path.endswith(patch_filepath)]
        if not xml_file_path_list:
            return
        xml_file_path = xml_file_path_list[0]
        xml_name = xml_file_path.split('/')[-1]
        xml_parse_result_new: mapper_parse.Mapper = self.xml_parse_results_new.get(xml_file_path)
        methods = xml_parse_result_new.result_maps + xml_parse_result_new.sqls + xml_parse_result_new.statements
        self._xml_method_diff_analyze(methods, diff_parse_obj['line_num_added'], diff_parse_obj['line_content_added'], xml_parse_result_new, xml_name, xml_file_path, self.commit_or_branch_new)
        xml_parse_result_old = self.xml_parse_results_old.get(xml_file_path)
        methods = xml_parse_result_old.result_maps + xml_parse_result_old.sqls + xml_parse_result_old.statements
        self._xml_method_diff_analyze(methods, diff_parse_obj['line_num_removed'], diff_parse_obj['line_content_removed'], xml_parse_result_old, xml_name, xml_file_path, self.commit_or_branch_old)

    # Step 4.1.1
    def _xml_method_diff_analyze(self, methods: list, line_num_list: list, line_content_list: list, xml_parse_result, xml_name, xml_file_path, commit_or_branch):
        namespace = xml_parse_result.namespace
        mapper_extend_dict = {
            'mapper_file_name': xml_name,
            'mapper_filepath': xml_file_path
        }
        for line_num in line_num_list:
            method_changed = [method for method in methods if self._is_line_num_in_xml_method_range(method, line_num)]
            method_changed_name = [method.name for method in method_changed]
            for method in method_changed:
                diff_content = line_content_list[line_num_list.index(line_num)]
                method_node_id = self.view.create_node_category(xml_name, method.name, method.type, constant.DIFF_TYPE_CHANGED, diff_content, xml_file_path, '', method.content, {})
                if method.type == constant.NODE_TYPE_MAPPER_STATEMENT:
                    mapper_extend_dict['method_node_id'] = method_node_id
                    self._add_to_need_analyze_obj_list('xml', namespace, None, method.name, commit_or_branch, mapper_extend_dict)
                    continue
                for statement in xml_parse_result.statements:
                    if statement.result_map in method_changed_name or statement.include_sql in method_changed_name:
                        statement_node_id = self.view.create_node_category(xml_name, statement.name, statement.type, constant.DIFF_TYPE_IMPACTED, '', xml_file_path, '', statement.content, {})
                        self.view.create_node_link(method_node_id, statement_node_id)
                        mapper_extend_dict['method_node_id'] = statement_node_id
                        self._add_to_need_analyze_obj_list('xml', namespace, None, statement.name, commit_or_branch, mapper_extend_dict)

    # Step 4.1.1.1
    def _is_line_num_in_xml_method_range(self, method, line_num):
        line_num_in_method = False
        if method.start <= line_num <= method.end:
            line_num_in_method = True
        return line_num_in_method

    # Step 4.2
    def _java_diff_analyze(self, patch_filepath: str, diff_parse_obj: dict):
        # new branch or commit
        class_db = self.sqlite.select_data(f'''SELECT * FROM class WHERE project_id = {self.project_id} and commit_or_branch = "{self.commit_or_branch_new}" and filepath LIKE "%{patch_filepath}"''')
        if class_db:
            class_db_obj = class_db[0]
            self._java_field_method_diff_analyze(class_db_obj, diff_parse_obj['line_num_added'], diff_parse_obj['line_content_added'], self.commit_or_branch_new)
        # old branch or commit
        if not self.commit_or_branch_old:
            return
        class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE project_id = {self.project_id} and commit_or_branch = "{self.commit_or_branch_old}" and filepath LIKE "%{patch_filepath}"')
        if class_db:
            class_db_obj = class_db[0]
            self._java_field_method_diff_analyze(class_db_obj, diff_parse_obj['line_num_removed'], diff_parse_obj['line_content_removed'], self.commit_or_branch_old)

    # Step 4.2.1
    def _java_field_method_diff_analyze(self, class_db: dict, line_num_list: list, line_content_list: list, commit_or_branch: str or None):
        if not commit_or_branch:
            return
        class_name = class_db['class_name']
        class_filepath = class_db['filepath']
        is_controller = class_db['is_controller']
        data_in_annotation = [annotation for annotation in json.loads(class_db['annotations']) if annotation['name'] in ['Data', 'Getter', 'Setter', 'Builder', 'NoArgsConstructor', 'AllArgsConstructor']]
        for line_num in line_num_list:
            diff_content = line_content_list[line_num_list.index(line_num)]
            fields_list = self.sqlite.select_data(f'SELECT * FROM field WHERE class_id = {class_db["class_id"]} AND start_line <={line_num} AND end_line >= {line_num} order by start_line asc limit 1')
            methods_list = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id = {class_db["class_id"]} AND start_line <={line_num} AND end_line >= {line_num} order by start_line asc limit 1')
            if fields_list:
                is_not_static_fields = [field for field in fields_list if field.get('is_static') == 'False']
                if is_not_static_fields and data_in_annotation:
                    self._add_to_need_analyze_obj_list('java', f'{class_db["package_name"]}.{class_name}', None, None, commit_or_branch, class_db)
                elif is_not_static_fields and not data_in_annotation:
                    field_method_name = []
                    for field in is_not_static_fields:
                        field_name = field['field_name']
                        field_name_capitalize = field_name[0].upper() + field_name[1:]
                        field_method_name += ['get' + field_name_capitalize, 'set' + field_name_capitalize, 'is' + field_name_capitalize]
                    field_method_name_str = '"' + '","'.join(field_method_name) + '"'
                    field_method_db = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id = {class_db["class_id"]} AND method_name in ({field_method_name_str})')
                    if field_method_db:
                        self._add_to_need_analyze_obj_list('java', f'{class_db["package_name"]}.{class_name}', None, None, commit_or_branch, class_db)
            for field_db in fields_list:
                node_id = self.view.create_node_category(class_name, field_db['field_name'], constant.NODE_TYPE_FIELD, constant.DIFF_TYPE_CHANGED, diff_content, class_filepath, field_db['documentation'], '', {})
                field_db['field_node_id'] = node_id
                self._add_to_need_analyze_obj_list('java', f'{class_db["package_name"]}.{class_name}', field_db['field_name'], None, commit_or_branch, field_db)
            for method_db in methods_list:
                node_extend_dict = {'is_api': False}
                if is_controller and method_db['is_api']:
                    node_extend_dict = {
                        'is_api': True,
                        'api_path': method_db['api_path']
                    }
                method_name_param = f'{method_db["method_name"]}({",".join([param["parameter_type"] for param in json.loads(method_db["parameters"])])})'
                node_id = self.view.create_node_category(class_name, method_name_param, constant.NODE_TYPE_METHOD, constant.DIFF_TYPE_CHANGED, diff_content, class_filepath, method_db['documentation'], method_db['body'], node_extend_dict)
                method_db['method_node_id'] = node_id
                self._add_to_need_analyze_obj_list('java', f'{class_db["package_name"]}.{class_name}', None, method_name_param, commit_or_branch, method_db)

    # Step 5
    def _impacted_analyze(self, need_analyze_obj: dict):
        file_type = need_analyze_obj['file_type']
        package_class = need_analyze_obj['package_class']
        commit_or_branch = need_analyze_obj['commit_or_branch']
        package_name = '.'.join(package_class.split('.')[0: -1])
        class_name = package_class.split('.')[-1]
        class_db_list = self.sqlite.select_data(f'SELECT * FROM class WHERE project_id = {self.project_id} and class_name="{class_name}" and package_name="{package_name}"')
        class_entity = self._get_right_class_entity(class_db_list, commit_or_branch)
        class_filepath = class_entity['filepath']
        class_id = class_entity["class_id"]
        # gengxin
        commit_or_branch = class_entity['commit_or_branch']
        is_controller = class_entity['is_controller']
        # todo 粗查，待细化
        if file_type == 'xml':
            method_name = need_analyze_obj['method_param']
            mapper_method_node_id = need_analyze_obj['method_node_id']
            impacted_methods = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id={class_id} and method_name="{method_name}"')
            if not impacted_methods:
                return
            for impacted_method in impacted_methods:
                node_extend_dict = {'is_api': False}
                if is_controller and impacted_method['is_api']:
                    node_extend_dict = {
                        'is_api': True,
                        'api_path': impacted_method['api_path']
                    }
                method_name_param = f'{impacted_method["method_name"]}({",".join([param["parameter_type"] for param in json.loads(impacted_method["parameters"])])})'
                impacted_method_node_id = self.view.create_node_category(class_name, method_name_param,
                                                                         constant.NODE_TYPE_METHOD, constant.DIFF_TYPE_IMPACTED,
                                                                         impacted_method['body'], class_filepath, impacted_method['documentation'],
                                                                         impacted_method['body'], node_extend_dict)
                self.view.create_node_link(mapper_method_node_id, impacted_method_node_id)
                extend_dict = {'method_node_id': impacted_method_node_id}
                extend_dict.update(impacted_method)
                self._add_to_need_analyze_obj_list('java', package_class, None, self._get_method_param_string(impacted_method), commit_or_branch, extend_dict)
        else:
            # analyze entity use
            entity_impacted_methods = []
            source_node_id = None
            if not need_analyze_obj.get('field_name') and not need_analyze_obj.get('method_param'):
                class_node_id = self.view.create_node_category(class_name, 'entity', constant.NODE_TYPE_CLASS, constant.DIFF_TYPE_IMPACTED, '', self.file_path, '', '', {})
                entity_impacted_methods = self._get_entity_invocation_in_methods_table(package_class)
                source_node_id = class_node_id
            elif need_analyze_obj.get('field_name'):
                annotations: list = json.loads(class_entity['annotations'])
                entity_impacted_methods = self._get_field_invocation_in_methods_table(package_class, need_analyze_obj, annotations, commit_or_branch)
                source_node_id = need_analyze_obj.get('field_node_id')
            elif need_analyze_obj.get('method_param'):
                method_param = need_analyze_obj.get('method_param')
                method_name: str = method_param.split('(')[0]
                method_node_id = need_analyze_obj.get('method_node_id')
                entity_impacted_methods = self._get_method_invocation_in_methods_table(package_class, method_param, commit_or_branch)
                method_db = self.sqlite.select_data(f'SELECT * FROM methods WHERE method_id = {need_analyze_obj.get("method_id")}')[0]
                is_override_method = 'Override' in method_db['annotations']
                if is_override_method and class_entity['extends_class']:
                    abstract_package_class = self._is_method_param_in_extends_package_class(method_param, class_entity['extends_class'], 'True', commit_or_branch)
                    if abstract_package_class:
                        extends_methods = self._get_method_invocation_in_methods_table(abstract_package_class, method_param, commit_or_branch)
                        for method in extends_methods:
                            method['class_id'] = class_id
                        entity_impacted_methods += extends_methods
                else:
                    class_method_db = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id = {class_id} and method_name = "{method_name}"')
                    if not class_method_db:
                        extends_package_class = self._is_method_param_in_extends_package_class(method_param, class_entity['extends_class'], 'False', commit_or_branch)
                        if extends_package_class:
                            extends_methods = self._get_method_invocation_in_methods_table(extends_package_class, method_param, commit_or_branch)
                            entity_impacted_methods += extends_methods
                source_node_id = method_node_id
            if not entity_impacted_methods:
                return
            self._handle_impacted_methods(entity_impacted_methods, source_node_id)

    def _is_method_param_in_extends_package_class(self, method_param, extends_package_class, is_abstract, commit_or_branch):
        method_name: str = method_param.split('(')[0]
        extends_package = '.'.join(extends_package_class.split('.')[0: -1])
        extends_class_name = extends_package_class.split('.')[-1]
        extends_class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE package_name = "{extends_package}" and class_name = "{extends_class_name}" and project_id = {self.project_id} and commit_or_branch = "{commit_or_branch}"')
        if not extends_class_db:
            return None
        extends_class_id = extends_class_db[0]['class_id']
        methods_db = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id = {extends_class_id} and method_name = "{method_name}" and is_abstract = "{is_abstract}"')
        if methods_db:
            return extends_package_class
        else:
            if extends_class_db[0]['extends_class']:
                return self._is_method_param_in_extends_package_class(method_param, extends_class_db[0]['extends_class'], is_abstract, commit_or_branch)
            else:
                return None

    def _get_extends_package_class(self, package_class):
        extends_package_class_list = []
        extends_package_class_db = self.sqlite.select_data(f'SELECT * FROM class WHERE project_id = {self.project_id} AND extends_class="{package_class}"')
        if extends_package_class_db:
            extends_package_class_list = [f'{class_item["package_name"]}.{class_item["class_name"]}' for class_item in extends_package_class_db]
            for extends_package_class in extends_package_class_list:
                extends_package_class_list += self._get_extends_package_class(extends_package_class)
        return extends_package_class_list

    # Step 5.1
    def _get_right_class_entity(self, class_db_list, commit_or_branch):
        right_class_entity = next((item for item in class_db_list if item.get("commit_or_branch") == commit_or_branch), None)
        if right_class_entity is None:
            right_class_entity = next((item for item in class_db_list if item.get("commit_or_branch") == self.commit_or_branch_new), None)
        return right_class_entity

    # Step 5.2
    def _get_entity_invocation_in_methods_table(self, package_class: str):
        return self.sqlite.select_data(f'''SELECT * FROM methods WHERE project_id = {self.project_id} AND json_extract(method_invocation_map, '$."{package_class}".entity') IS NOT NULL''')

    # Step 5.3
    def _get_field_invocation_in_methods_table(self, package_class, field_obj, annotations, commit_or_branch):
        is_static = field_obj['is_static']
        field_name = field_obj['field_name']
        if not is_static:
            # todo static maybe has bug
            field_methods_set = set()
            field_name_capitalize = field_name[0].upper() + field_name[1:]
            for annotation in annotations:
                annotation_name = annotation.get('name')
                if annotation_name == 'Data':
                    field_methods_set.add(f'get{field_name_capitalize}(')
                    field_methods_set.add(f'set{field_name_capitalize}(')
                elif annotation_name == 'Getter':
                    field_methods_set.add(f'get{field_name_capitalize}(')
                elif annotation_name == 'Setter':
                    field_methods_set.add(f'set{field_name_capitalize}(')
                else:
                    continue
            if not field_methods_set:
                return []
            json_extract_sql_list = []
            for field_method in field_methods_set:
                sql_part = f'''json_extract(method_invocation_map, '$."{package_class}".methods.keys(@.startsWith("{field_method}"))') IS NOT NULL'''
                json_extract_sql_list.append(sql_part)
            sql = f'SELECT * FROM methods WHERE project_id = {self.project_id} AND (' + ' OR '.join(json_extract_sql_list) + ')'
        else:
            sql = f'''SELECT * FROM methods WHERE project_id = {self.project_id} AND json_extract(method_invocation_map, '$."{package_class}".fields.{field_name}') IS NOT NULL'''
        methods = self.sqlite.select_data(sql)
        if not methods:
            return []
        class_ids = [str(method['class_id']) for method in methods]
        class_sql = f'SELECT * FROM class WHERE class_id in ({", ".join(class_ids)}) and commit_or_branch ="{commit_or_branch}"'
        class_db = self.sqlite.select_data(class_sql)
        class_db_id = [class_item['class_id'] for class_item in class_db]
        return [method for method in methods if method['class_id'] in class_db_id]

    # Step 5.4
    def _get_method_invocation_in_methods_table(self, package_class, method_param, commit_or_branch):
        all_possible_method_param_type_list = self._gen_all_possible_method_param_list(method_param)
        json_extract_sql_list = []
        for param_type in all_possible_method_param_type_list:
            sql_part = f'''json_extract(method_invocation_map, '$."{package_class}".methods."{param_type}"') IS NOT NULL'''
            json_extract_sql_list.append(sql_part)
        if len(json_extract_sql_list) > 1000:
            json_extract_sql_list = json_extract_sql_list[0: 995]
        sql = f'SELECT * FROM methods WHERE project_id = {self.project_id} AND (' + ' OR '.join(json_extract_sql_list) + ')'
        logging.info(f'{package_class} {method_param} invocation sql: {sql}')
        methods = self.sqlite.select_data(sql)
        class_ids = [str(method['class_id']) for method in methods]
        class_sql = f'SELECT * FROM class WHERE class_id in ({", ".join(class_ids)}) and commit_or_branch ="{commit_or_branch}"'
        class_db = self.sqlite.select_data(class_sql)
        class_db_id = [class_item['class_id'] for class_item in class_db]
        return [method for method in methods if method['class_id'] in class_db_id]

    # Step 5.4.1
    def _gen_all_possible_method_param_list(self, method_param):
        method_param_list = []
        method_name = method_param.split('(')[0]
        param_type_str = method_param.split('(')[1].split(')')[0]
        param_type_list = param_type_str.split(',')
        if not param_type_list:
            return method_param_list
        all_possible_method_param_list = self._replace_with_null(param_type_list)
        for param_type_list in all_possible_method_param_list:
            method_param_list.append(f'{method_name}({",".join(param_type_list)})')
        return method_param_list

    def _replace_with_null(self, lst):
        # lst = [item.split('<')[0].replace('<', '').replace('>', '') for item in lst]
        results = []
        lowercase_indices = [i for i, s in enumerate(lst) if s and s[0].islower()]  # 首字母小写的元素索引

        for num_replacements in range(len(lst) + 1):  # 遍历替换0到所有元素的情况
            replaceable_indices = [i for i in range(len(lst)) if i not in lowercase_indices]  # 可替换的元素索引
            for replace_indices in combinations(replaceable_indices, num_replacements):  # 所有可能的替换索引组合
                new_lst = lst[:]  # 创建原始列表的副本
                for idx in replace_indices:
                    if new_lst[idx].lower() not in constant.JAVA_BASIC_TYPE:
                        new_lst[idx] = constant.PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
                    else:
                        new_lst[idx] = 'null'
                        new_lst2 = new_lst[:]
                        new_lst2[idx] = constant.PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
                        results.append(new_lst2)
                        self._replace_extends_class(new_lst2, results)
                results.append(new_lst)
                self._replace_extends_class(new_lst, results)
        return results

    def _replace_extends_class(self, new_lst, results):
        for i in range(0, len(new_lst)):
            if new_lst[i].lower() in constant.JAVA_BASIC_TYPE \
                    or new_lst[i] == constant.PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN \
                    or new_lst[i] == 'null':
                continue
            extends_package_class_list = self._get_extends_package_class(new_lst[i])
            for extends_package_class in extends_package_class_list:
                result_item = [item for item in new_lst]
                result_item[i] = extends_package_class
                results.append(result_item)

    # Step 5.4.1.1
    def _replace_with_unknown(self, data: list):
        # data = [item.split('<')[0].replace('<', '').replace('>', '') for item in data]
        results = [data]
        for i in range(0, len(data)):
            result_item = [item for item in data]
            if data[i].lower() not in constant.JAVA_BASIC_TYPE:
                result_item[i] = constant.PARAMETER_TYPE_METHOD_INVOCATION_UNKNOWN
            elif data[i][0].isupper():
                result_item[i] = 'null'
            results.append(result_item)
        for i in range(0, len(data)):
            if data[i].lower() in constant.JAVA_BASIC_TYPE:
                continue
            extends_package_class_list = self._get_extends_package_class(data[i])
            for extends_package_class in extends_package_class_list:
                result_item = [item for item in data]
                result_item[i] = extends_package_class
                results.append(result_item)
        return results

    # Step 5.5
    def _get_method_param_string(self, method_db: dict):
        method_name: str = method_db['method_name']
        params: list = json.loads(method_db['parameters'])
        params_type_list = [param['parameter_type'] for param in params]
        return f'{method_name}({",".join(params_type_list)})'

    # Step 5.9
    def _handle_impacted_methods(self, impacted_methods: list, source_node_id):
        for impacted_method in impacted_methods:
            node_extend_dict = {'is_api': False}
            if impacted_method.get('is_api') == 'True':
                node_extend_dict = {
                    'is_api': True,
                    'api_path': impacted_method['api_path']
                }
            class_id = impacted_method['class_id']
            class_entity = self.sqlite.select_data(f'SELECT * FROM class WHERE class_id={class_id}')[0]
            class_name = class_entity['class_name']
            package_name = class_entity['package_name']
            package_class = f'{package_name}.{class_name}'
            commit_or_branch = class_entity['commit_or_branch']
            class_filepath = class_entity['filepath']
            method_name_param = f'{impacted_method["method_name"]}({",".join([param["parameter_type"] for param in json.loads(impacted_method["parameters"])])})'
            impacted_method_node_id = self.view.create_node_category(class_name, method_name_param, constant.NODE_TYPE_METHOD, constant.DIFF_TYPE_IMPACTED, impacted_method['body'], class_filepath, impacted_method['documentation'], impacted_method['body'], node_extend_dict)
            self.view.create_node_link(source_node_id, impacted_method_node_id)
            extend_dict = {'method_node_id': impacted_method_node_id, 'class_filepath': class_filepath}
            extend_dict.update(impacted_method)
            self._add_to_need_analyze_obj_list('java', package_class, None, self._get_method_param_string(impacted_method), commit_or_branch, extend_dict)

    def _add_to_need_analyze_obj_list(self, file_type: str, package_class: str, field_name: str or None, method_param: str or None, commit_or_branch: str, mapper_extend_dict: dict):
        need_analyze_entity: dict = {
            'file_type': file_type,
            'package_class': package_class,
            'field_name': field_name,
            'method_param': method_param,
            'commit_or_branch': commit_or_branch
        }
        if need_analyze_entity not in self.need_analyze_obj_list:
            need_analyze_entity.update(mapper_extend_dict)
            self.need_analyze_obj_list.append(need_analyze_entity)

    def _draw_and_write_result(self):
        if self.view.nodes:
            self.view.draw_graph(1200, 600)
        logging.info(f'Analyze success, generating cci result file......')
        result = {
            'nodes': self.view.nodes,
            'links': self.view.links,
            'categories': self.view.categories,
            'impacted_api_list': [node['api_path'] for node in self.view.nodes if node.get('is_api')]
        }
        print(json.dumps(result), flush=True)
        print(f'Impacted api list: {result["impacted_api_list"]}', flush=True)
        with open(self.cci_filepath, 'w') as w:
            w.write(json.dumps(result))
        logging.info(f'Generating cci result file success, location: {self.cci_filepath}')

    def _start_analysis_diff_and_impact(self):
        for patch_path, patch_obj in self.diff_parse_map.items():
            self._diff_analyze(patch_path, patch_obj)

        # 遍历列表
        for obj in self.need_analyze_obj_list:
            if obj not in self.analyzed_obj_set:  # 判断对象是否已分析过
                self.analyzed_obj_set.append(obj)  # 标记为已分析过
                self._impacted_analyze(obj)  # 处理对象,返回新增对象列表

        self._draw_and_write_result()
        t2 = datetime.datetime.now()
        try:
            logging.info(f'Analyze done, remove occupy, others can analyze now')
            os.remove(os.path.join(self.file_path, 'Occupy.ing'))
        finally:
            pass
        logging.info(f'Analyze done, spend: {t2 - self.t1}')

    def analyze_two_commit(self, branch, commit_first, commit_second):
        logging.info('*' * 10 + 'Analyze start' + '*' * 10)
        self.branch_name = branch
        self.commit_or_branch_new = commit_first[0: 7] if len(commit_first) > 7 else commit_first
        self.commit_or_branch_old = commit_second[0: 7] if len(commit_second) > 7 else commit_second

        self.project_name = self.git_url.split('/')[-1].split('.git')[0]
        self.file_path = os.path.join(project_path, self.project_name)

        self.project_id = self.sqlite.add_project(self.project_name, self.git_url, self.branch_name, self.commit_or_branch_new, self.commit_or_branch_old)
        # 已有分析结果
        self.cci_filepath = os.path.join(self.file_path, f'{self.commit_or_branch_old}..{self.commit_or_branch_new}.cci')
        self._can_analyze(self.file_path, self.cci_filepath)

        # 无此项目, 先clone项目
        if not os.path.exists(self.file_path):
            logging.info(f'Cloning project: {self.git_url}')
            os.system(f'git clone -b {self.branch_name} {self.git_url} {self.file_path}')

        self._occupy_project()
        self.diff_parse_map = self._get_diff_parse_map(self.file_path, self.branch_name, self.commit_or_branch_new, self.commit_or_branch_old)

        self.xml_parse_results_new, self.xml_parse_results_old = self._parse_project(self.file_path, self.commit_or_branch_new, self.commit_or_branch_old)

        self._start_analysis_diff_and_impact()

    def analyze_class_method(self, branch, commit_id, package_class, method_nums):
        logging.info('*' * 10 + 'Analyze start' + '*' * 10)
        package_class = package_class.replace("\\", "/")
        self.branch_name = branch
        self.commit_or_branch_new = commit_id
        self.commit_or_branch_new = self.commit_or_branch_new[0: 7] if len(self.commit_or_branch_new) > 7 else self.commit_or_branch_new
        self.project_name = self.git_url.split('/')[-1].split('.git')[0]
        self.file_path = os.path.join(project_path, self.project_name)

        project_id = self.sqlite.add_project(self.project_name, self.git_url, self.branch_name, self.commit_or_branch_new, f'{package_class}.{method_nums}')
        class_name = package_class.split("/")[-1].replace('.java', '')
        cci_path = f'{branch}_{commit_id}_{class_name}_{method_nums}.cci'
        self.cci_filepath = os.path.join(self.file_path, cci_path)
        self._can_analyze(self.file_path, self.cci_filepath)

        # 无此项目, 先clone项目
        if not os.path.exists(self.file_path):
            logging.info(f'Cloning project: {self.git_url}')
            os.system(f'git clone -b {self.branch_name} {self.git_url} {self.file_path}')
        self._occupy_project()

        logging.info('Git pull project to HEAD')
        os.system(f'cd {self.file_path} && git checkout {branch} && git pull')
        time.sleep(1)

        self.xml_parse_results_new, self.xml_parse_results_old = self._parse_project(self.file_path, self.commit_or_branch_new, None)

        if not method_nums:
            method_nums_all = []
            # todo
            class_db = self.sqlite.select_data('SELECT * FROM class WHERE project_id = ' + str(project_id) + ' and filepath LIKE "%' + package_class + '"')
            if not class_db:
                logging.error(f'Can not find {package_class} in db')
            class_id = class_db[0]['class_id']
            field_db = self.sqlite.select_data(f'SELECT * FROM field WHERE class_id = {class_id}')
            method_nums_all += [field['start_line'] for field in field_db]
            method_db = self.sqlite.select_data(f'SELECT * FROM methods WHERE class_id = {class_id}')
            method_nums_all += [method['start_line'] for method in method_db]
        else:
            method_nums_all = [int(num) for num in method_nums.split(',')]

        self.diff_parse_map[package_class] = {
            'line_num_added': method_nums_all,
            'line_content_added': method_nums_all,
            'line_num_removed': [],
            'line_content_removed': []
        }

        self._start_analysis_diff_and_impact()



import src.jcci.constant as constant
from collections import deque

def max_relationship_length(relationships):
    # 构建邻接列表
    graph = {}
    for relationship in relationships:
        source = relationship['source']
        target = relationship['target']
        if source not in graph:
            graph[source] = []
        if target not in graph:
            graph[target] = []
        graph[source].append(target)

    # BFS遍历计算每个节点到起点的最长路径长度
    longest_paths = {node: 0 for node in graph.keys()}
    graph_keys = [node for node in graph.keys()]
    longest_paths[graph_keys[0]] = 0
    queue = deque([(graph_keys[0], 0)])
    while queue:
        node, path_length = queue.popleft()
        if not graph.get(node) and not queue and graph_keys.index(node) + 1 < len(graph_keys):
            next_node = graph_keys[graph_keys.index(node) + 1]
            next_node_path_length = longest_paths[next_node]
            queue.append((next_node, next_node_path_length))
            continue
        for neighbor in graph.get(node, []):
            if path_length + 1 > longest_paths[neighbor]:
                longest_paths[neighbor] = path_length + 1
                queue.append((neighbor, path_length + 1))
    return longest_paths

class Graph(object):

    def __init__(self):
        self.nodes = []
        self.links = []
        self.categories = []
        self.node_index_init = 0

    def create_node_category(self, class_or_xml, name, type, diff_type, diff_content, file_path, documentation, body, extend_dict: dict):
        category = {
            'name': class_or_xml
        }
        if category not in self.categories:
            self.categories.append(category)
        if class_or_xml == name:
            return
        category_id = self.categories.index(category)
        node_create = {
            'category': category_id,
            'id': str(self.node_index_init),
            'name': class_or_xml + '.' + name,
            'type': type,
            'diff_type': [diff_type],
            'file_path': file_path,
        }
        if diff_type == constant.DIFF_TYPE_CHANGED:
            node_create.update({
                'diff_content': diff_content,
                'documentation': documentation,
                'body': body,
            })
        node_exist = [node for node in self.nodes if node['name'] == node_create['name'] and node['type'] == type and node['file_path'] == file_path]
        if node_exist:
            node_create: dict = node_exist[0]
            self.nodes.remove(node_create)
            if diff_type not in node_create['diff_type']:
                node_create['diff_type'].append(diff_type)
            node_create.update(extend_dict)
            self.nodes.append(node_create)
        else:
            node_create.update(extend_dict)
            self.nodes.append(node_create)
            self.node_index_init += 1
        return node_create['id']

    def create_node_link(self, source_node_id, target_node_id):
        if source_node_id is None or target_node_id is None:
            return
        if source_node_id == target_node_id:
            return
        link = {
            'source': source_node_id,
            'target': target_node_id
        }
        if link not in self.links:
            self.links.append(link)

    def draw_graph(self, canvas_width, canvas_height):
        # 每个类别区域划分的行数
        all_node = []
        result = max_relationship_length(self.links)
        changed_nodes = [node for node in self.nodes if 'changed' in node['diff_type']]
        impacted_nodes = [node for node in self.nodes if 'changed' not in node['diff_type']]
        for changed_node in changed_nodes:
            changed_node['x'] = 100
            changed_node['y'] = (changed_nodes.index(changed_node) + 1) * (canvas_height / len(changed_nodes))
            changed_node['symbolSize'] = 20
            changed_node['label'] = {
                'show': True,
                'formatter': changed_node["name"].split("(")[0]
            }
            tooltip = f'{changed_node["name"].split("(")[0]}<br>[Changed]{changed_node["diff_content"]}'
            if changed_node.get('is_api'):
                tooltip = tooltip + f'<br>[API]{changed_node["api_path"]}'
            changed_node['tooltip'] = {
                'show': True,
                'position': 'right',
                'formatter': tooltip
            }
            all_node.append(changed_node)
        max_link_count = max([value for key, value in result.items()])
        count_node_result = {}
        for key, value in result.items():
            value = str(value)
            if value not in count_node_result:
                count_node_result[value] = []
            count_node_result[value].append(key)
        for impacted_node in impacted_nodes:
            path_level = result.get(impacted_node['id'], 0)
            level_node_list = count_node_result.get(str(path_level), [impacted_node['id']])
            level_node_index = level_node_list.index(impacted_node['id']) if impacted_node['id'] in level_node_list else 1
            impacted_node['x'] = 100 + ((canvas_width - 100) / max_link_count) * (path_level + 1)
            impacted_node['y'] = (canvas_height / len(count_node_result[str(path_level)])) * level_node_index
            impacted_node['label'] = {
                'show': True,
                'formatter': impacted_node["name"].split("(")[0]
            }
            if impacted_node.get('is_api'):
                tooltip = f'{impacted_node["name"].split("(")[0]}<br>[API]{impacted_node["api_path"]}'
                impacted_node['tooltip'] = {
                    'show': True,
                    'position': 'right',
                    'formatter': tooltip
                }
            all_node.append(impacted_node)
        self.nodes = all_node

if __name__ == '__main__':
    data = [{'source': '62', 'target': '61'}, {'source': '63', 'target': '64'}, {'source': '63', 'target': '65'}, {'source': '63', 'target': '66'}, {'source': '63', 'target': '67'}, {'source': '9', 'target': '8'}, {'source': '68', 'target': '25'}, {'source': '21', 'target': '18'}, {'source': '21', 'target': '19'}, {'source': '26', 'target': '25'}, {'source': '27', 'target': '25'}, {'source': '72', 'target': '5'}, {'source': '72', 'target': '8'}, {'source': '72', 'target': '9'}, {'source': '73', 'target': '18'}, {'source': '73', 'target': '19'}, {'source': '73', 'target': '20'}, {'source': '73', 'target': '21'}, {'source': '74', 'target': '17'}, {'source': '74', 'target': '54'}, {'source': '75', 'target': '76'}, {'source': '75', 'target': '3'}, {'source': '75', 'target': '77'}, {'source': '75', 'target': '5'}, {'source': '75', 'target': '78'}, {'source': '75', 'target': '79'}, {'source': '75', 'target': '6'}, {'source': '75', 'target': '80'}, {'source': '75', 'target': '8'}, {'source': '75', 'target': '9'}, {'source': '75', 'target': '81'}, {'source': '46', 'target': '47'}, {'source': '46', 'target': '60'}, {'source': '47', 'target': '61'}, {'source': '48', 'target': '47'}, {'source': '49', 'target': '46'}, {'source': '50', 'target': '46'}, {'source': '83', 'target': '17'}, {'source': '83', 'target': '18'}, {'source': '54', 'target': '17'}, {'source': '55', 'target': '18'}, {'source': '56', 'target': '55'}, {'source': '57', 'target': '55'}, {'source': '58', 'target': '56'}, {'source': '58', 'target': '57'}, {'source': '59', 'target': '54'}, {'source': '59', 'target': '55'}, {'source': '59', 'target': '56'}, {'source': '59', 'target': '57'}, {'source': '64', 'target': '66'}, {'source': '65', 'target': '64'}]
    bb = max_relationship_length(data)

